/// <reference types="node" />
import { Keypair, Cluster, PublicKey } from '@solana/web3.js';
import { ClusterType } from '@identity.com/sol-did-client';
import { SignCallback } from './wallet';
import { AddressBook } from './UserDetails';
export declare const debug: any;
export declare type PrivateKey = number[] | string | Buffer | Uint8Array;
export declare type PublicKeyBase58 = string;
export declare const privateKeyIsArray: (privateKey: PrivateKey) => privateKey is number[];
export declare const privateKeyIsString: (privateKey: PrivateKey) => privateKey is string;
export declare const privateKeyIsBuffer: (privateKey: PrivateKey) => privateKey is Buffer;
export declare const privateKeyIsUint8Array: (privateKey: PrivateKey) => privateKey is Uint8Array;
/**
 * Create a Solana keypair object from an x25519 private key
 * @param privateKey
 */
export declare const makeKeypair: (privateKey: PrivateKey) => Keypair;
/**
 * Given a private key on the x25519 curve, get its public key
 * @param privateKey
 */
export declare const getPublicKey: (privateKey: PrivateKey) => PublicKey;
declare type EncodedKeyPair = {
    secretKey: string;
    publicKey: string;
};
export declare const generateKeypair: () => EncodedKeyPair;
export declare type ExtendedCluster = Cluster | 'localnet';
export declare const getClusterEndpoint: (cluster?: "devnet" | "testnet" | "mainnet-beta" | "localnet" | undefined) => string;
declare type KeyMaterial = Keypair | PrivateKey | PublicKeyBase58 | PublicKey;
export declare const toSolanaKeyMaterial: (k: KeyMaterial) => Keypair | PublicKey;
export declare type SolanaRequest = {
    cluster?: ExtendedCluster;
};
export declare type TransactionRequest = SolanaRequest & {
    payer: KeyMaterial;
    signCallback?: SignCallback;
};
export declare type CreateRequest = TransactionRequest & {
    owner?: KeyMaterial;
    name: string;
};
export declare type CreateDirectRequest = TransactionRequest & {
    owner?: KeyMaterial;
    inviteeDID: string;
};
export declare type CloseRequest = TransactionRequest & {
    ownerDID: string;
    signer?: PrivateKey;
};
export declare type PostRequest = TransactionRequest & {
    senderDID?: string;
    signer: PrivateKey;
    channel: PublicKeyBase58;
    message: string;
};
export declare type PostDirectRequest = TransactionRequest & {
    senderDID?: string;
    signer: PrivateKey;
    recipientDID: string;
    message: string;
};
export declare type ReadRequest = SolanaRequest & {
    memberDID?: string;
    member?: PublicKeyBase58;
    decryptionKey: PrivateKey;
    channel: PublicKeyBase58;
};
export declare type GetRequest = SolanaRequest & {
    ownerDID?: string;
    owner?: PublicKeyBase58;
    decryptionKey?: PrivateKey;
    channel: PublicKeyBase58;
};
export declare type GetDirectRequest = SolanaRequest & {
    ownerDID?: string;
    partnerDID: string;
    owner?: PublicKeyBase58;
    decryptionKey?: PrivateKey;
};
export declare type AddKeyRequest = TransactionRequest & {
    ownerDID?: string;
    signer: KeyMaterial;
    newKey: PublicKeyBase58;
    keyIdentifier: string;
    channelsToUpdate: PublicKeyBase58[];
};
export declare type AddToChannelRequest = TransactionRequest & {
    ownerDID?: string;
    inviteeDID: string;
    decryptionKey: PrivateKey;
    channel: PublicKeyBase58;
};
export declare type CreateDIDRequest = TransactionRequest & {
    owner?: KeyMaterial;
    alias?: string;
};
export declare type GetDIDRequest = SolanaRequest & {
    owner: PublicKeyBase58;
};
export declare type CreateUserDetailsRequest = TransactionRequest & {
    ownerDID?: string;
    owner?: KeyMaterial;
    alias: string;
    size?: number;
};
export declare type UpdateUserDetailsRequest = TransactionRequest & {
    ownerDID?: string;
    owner?: KeyMaterial;
    alias?: string;
    addressBook?: AddressBook;
};
export declare type GetUserDetailsRequest = SolanaRequest & {
    did: string;
};
export declare const didToPublicKey: (did: string) => PublicKey;
export declare const currentCluster: (cluster?: "devnet" | "testnet" | "mainnet-beta" | "localnet" | undefined) => ClusterType;
export declare const arrayOf: <T>(...things: (T | undefined)[]) => T[];
export declare const isKeypair: (k: KeyMaterial) => k is Keypair;
export declare const isPublicKey: (k: KeyMaterial) => k is PublicKey;
export declare const pubkeyOf: (k: Keypair | PublicKey) => PublicKey;
export declare const isString: (value: any) => value is string;
export {};
